from .direction import Direction


import logging
logging.basicConfig(level=logging.DEBUG, format='(%(threadName)-9s) %(message)s',)


''' control robot positioning and drawing '''
class RobotPosition():
        
    move_count = 0   
    maze = None
    

    def __init__( self, level, **kwargs ):                
                
      self.level = level.grid_base
      self.grid  = level.draw_grid
      
      if hasattr(level, 'maze'):
        self.maze = level.maze
        
      # the position in grid cells
      self.x_cell = 0
      self.y_cell = 0
      
      # the position in pixels                        
      self.x = 0
      self.y = 0     
      
      self.step = 4
      self.robot_size = 64
              
      if self.maze is None:
          self.x_size = self.grid.width_pixels
          self.y_size = self.grid.height_pixels
      else:
          x,y = self.maze.dimensions()
          self.x_size = x * self.robot_size
          self.y_size = y * self.robot_size                                   

      # baby robot's initial position     
      self.initial_position = kwargs.get('initial_pos',self.level.start)
      self.set_cell_position(self.initial_position)        


    def get_cell_position(self):
      ''' get the current position in grid coords '''
      return self.x//self.robot_size, self.y//self.robot_size


    def set_cell_position(self, *args):
      ''' set the robot position in grid coords '''      
      # convert from grid position to pixels
      if len(args) == 1:    
        self.x,self.y = self.grid.grid_to_pixels(args[0])
        self.x_cell = args[0][0]
        self.y_cell = args[0][1]
      elif len(args) == 2: 
        self.x,self.y = self.grid.grid_to_pixels([args[0],args[1]])     
        self.x_cell = args[0]
        self.y_cell = args[1]        
                          
                
    def move(self,new_x,new_y):
        ' move from the current position to the specified position '                

        while (self.x_cell < (self.level.width-1)) and (self.x_cell < new_x):          
          self.move_direction(Direction.East)
          self.x_cell += 1

        while (self.x_cell > 0) and (self.x_cell > new_x):          
          self.move_direction(Direction.West)      
          self.x_cell -= 1

        while (self.y_cell < (self.level.height-1)) and (self.y_cell < new_y):          
          self.move_direction(Direction.South)        
          self.y_cell += 1

        while (self.y_cell > 0) and (self.y_cell > new_y):          
          self.move_direction(Direction.North)                     
          self.y_cell -= 1


    def test_for_valid_move( self, direction ):
        ''' test if a move can actually be made in the specified direction '''
                
        if direction == Direction.Stay:
          return False

        cell = None
        if self.maze is not None:
            x, y = self.get_cell_position()
            cell = self.maze.cell_at( x, y )            
            if direction == Direction.North and cell.walls['N']: return False
            if direction == Direction.South and cell.walls['S']: return False
            if direction == Direction.East and cell.walls['E']: return False
            if direction == Direction.West and cell.walls['W']: return False        

        return True                        
        
        
    def move_East(self):        
        if self.x < (self.x_size - self.robot_size):
            self.x += self.step
            
    def move_West(self):
        if self.x > 0: 
            self.x -= self.step            
            
    def move_North(self):
        if self.y > 0:           
            self.y -= self.step
            
    def move_South(self):
        if self.y < (self.y_size - self.robot_size):
            self.y += self.step


import gym


class BabyRobotEnv_v0(gym.Env):
    
    def __init__(self):
        super().__init__()
        pass

    def step(self, action):        
        state = 1    
        reward = -1            
        done = True
        info = {}
        return state, reward, done, info

    def reset(self):
        state = 0
        return state
  
    def render(self,mode):
        pass




from .direction import Direction

class Arrows():
  ''' draw arrows to show policies '''
  
  a_width = 6
  a_height = 7
  line_length = 14
  
  def __init__(self,cell_pixels,padding,length=14,width=6,height=7):
    self.cell_pixels = cell_pixels
    self.padding = padding
    self.calculate_positions()
    self.line_length = length
    self.a_width = width
    self.a_height = height
    
  def set_center(self,x,y):
    ''' set the center of the tile '''
    self.cx = x + self.start
    self.cy = y + self.start     
    
  def calculate_positions(self):
    ''' calculate the number of pixels to center of a square '''
    self.start = self.cell_pixels//2 - self.padding
    
  def draw_horz_line(self,canvas,shift):
    canvas.begin_path()     
    canvas.move_to(self.cx + shift,self.cy)
    canvas.line_to(self.cx + shift + self.line_length,self.cy)     
    canvas.stroke()     
    
  def draw_vert_line(self,canvas,shift):
    canvas.begin_path()     
    canvas.move_to(self.cx,self.cy + shift)
    canvas.line_to(self.cx,self.cy + shift + self.line_length )      
    canvas.stroke()              
    
  def draw_vert_arrow(self,canvas,y1,y2):
    canvas.begin_path()
    canvas.move_to(self.cx, y2)
    canvas.line_to(self.cx + self.a_width, y1)
    canvas.line_to(self.cx - self.a_width, y1)
    canvas.fill() 
    
  def draw_horz_arrow(self,canvas,x1,x2):
    canvas.begin_path()
    canvas.move_to(x2, self.cy)
    canvas.line_to(x1, self.cy + self.a_width)
    canvas.line_to(x1, self.cy - self.a_width)
    canvas.fill()            
    
  def draw_arrow(self,canvas,x,y,direction):
    
    if direction == Direction.North: 
      y1 = self.cy - self.line_length + 1
      y2 = y1 - self.a_height + 1
      self.draw_vert_line(canvas,-self.line_length+1)
      self.draw_vert_arrow(canvas,y1,y2)
      
    if direction == Direction.South:       
      y1 = self.cy + self.line_length - 1
      y2 = y1 + self.a_height - 1       
      self.draw_vert_line(canvas,-1)   
      self.draw_vert_arrow(canvas,y1,y2)
      
    if direction == Direction.East:  
      x1 = self.cx + self.line_length - 1
      x2 = x1 + self.a_height - 1      
      self.draw_horz_line(canvas,-1)
      self.draw_horz_arrow(canvas,x1,x2)     
      
    if direction == Direction.West:  
      x1 = self.cx - self.line_length + 1
      x2 = x1 - self.a_height + 1     
      self.draw_horz_line(canvas,-self.line_length+1)     
      self.draw_horz_arrow(canvas,x1,x2)           
    
  def draw(self,canvas,x,y,directions,color='#000',center_width = 36):   
    ''' draw arrows in the specified directions onto the supplied canvas '''

    # if the directions are not a list then they should be a bitfield, set for each direction
    if type(directions) is not list:
      # convert the direction value into a list of directions
      directions = Direction.get_list( directions )     

    canvas.line_width = 3
    canvas.fill_style = color
    canvas.stroke_style = color      
    
    self.set_center(x,y)    
    for direction in directions:          
      self.draw_arrow(canvas,x,y,direction)  
    
    # if a center width is defined then clear this area    
    if center_width > 0: 
      center_start = center_width//2
      canvas.clear_rect(self.cx-center_start,self.cy-center_start,center_width,center_width) 



import numpy as np
from .baby_robot_env_v1 import BabyRobotEnv_v1
from .lib.actions import Actions


class BabyRobotEnv_v2( BabyRobotEnv_v1 ):

  metadata = {'render_modes': ['human']}
  
  def __init__(self, **kwargs):
      super().__init__(**kwargs)
      
      # the start and end positions in the grid
      # - by default these are the top-left and bottom-right respectively
      self.start = kwargs.get('start',[0,0])       
      self.end = kwargs.get('end',[self.max_x,self.max_y])        
      
      # Baby Robot's initial position
      # - by default this is the grid start 
      self.initial_pos = kwargs.get('initial_pos',self.start)  

      # Baby Robot's position in the grid
      self.x = self.initial_pos[0]
      self.y = self.initial_pos[1]
      
      
  def take_action(self, action):
      ''' apply the supplied action '''
      
      # move in the direction of the specified action
      if   action == Actions.North: self.y -= 1
      elif action == Actions.South: self.y += 1
      elif action == Actions.West:  self.x -= 1
      elif action == Actions.East:  self.x += 1    
      
      # make sure the move stays on the grid
      if self.x < 0: self.x = 0
      if self.y < 0: self.y = 0
      if self.x > self.max_x: self.x = self.max_x
      if self.y > self.max_y: self.y = self.max_y        

        
  def step(self, action): 

      # take the action and update the position
      self.take_action(action)      
      obs = np.array([self.x,self.y])
            
      # set the 'done' flag if we've reached the exit
      done = (self.x == self.end[0]) and (self.y == self.end[1])
      
      # get -1 reward for each step
      # - except at the terminal state which has zero reward
      reward = 0 if done else -1
        
      info = {}
      return obs, reward, done, info     


  def render(self, mode='human', action=0, reward=0 ):
      if mode == 'human':
        print(f"{Actions(action): <5}: ({self.x},{self.y}) reward = {reward}") 
      else:
        super().render(mode=mode) # just raise an exception      


import os
from enum import IntEnum
from random import uniform
from math import pi
import json

from ipycanvas import MultiCanvas, Canvas, hold_canvas
from ipywidgets import Image


from babyrobot.envs.lib import GridBase
from babyrobot.envs.lib import Arrows
from babyrobot.envs.lib import Direction



class Level(IntEnum): 
    Base     = 0
    Grid     = 1
    Underlay = 2    
    Robot    = 3
    Overlay  = 4



class DrawGrid():

  num_canvases = 5           # number of canvases/layers
  cell_pixels = 64           # pixel dimensions of a grid square     
  padding = 2                # padding around the cells
  wall_width = 4             # the width of maze walls  
  border_width = 5           # the width of the outside border
  side_panel = None          # by default there's no side info panel
  bottom_panel = None        # by default there's no bottom info panel

  base_color = 'orange'      # color of the grid base layer
  grid_color = '#777'        # grid line color
  start_color = '#ed1818'    # color of start square
  start_text_color = '#fff'  
  exit_color = 'green'       # color of the exit square
  exit_text_color = '#fff'
  border_color = 'black'     # color of the outer border
  wall_color = 'black'       # color of the walls


  def __init__(self, gridbase: GridBase, **kwargs: dict):

    self.grid = gridbase

    self.show_start_text = kwargs.get('show_start_text',True)
    self.show_end_text = kwargs.get('show_end_text',True)        

    # setup the grid properties
    self.set_properties(kwargs.get('grid',None))  

    # setup any information items
    self.add_compass = kwargs.get('add_compass',False)
    self.side_panel = kwargs.get('side_panel',None)    
    self.bottom_panel = kwargs.get('bottom_panel',None)            

    # load the image used to draw puddles
    self.load_puddle_sprite() 

    # create the set of canvases for drawing
    self.create_canvases()    
    self.draw_level()  


  '''
      Setup Functions
  '''

  def set_properties( self, grid_props: dict ):
    ''' setup the grid draw properties '''    
    
    if grid_props is not None:

      # first test if a theme is specified
      theme = grid_props.get('theme',None)
      if theme is not None:       
        # look for packaged themes
        theme_path = os.path.join(self.grid.working_directory,f'themes/{theme}.json')
        if os.path.exists(theme_path):         
          with open(theme_path) as json_file:
            grid_props = json.load(json_file)        
        else:
          # look for user defined themes      
          theme_path = os.path.join(os.getcwd(),f'themes/{theme}.json')
          if os.path.exists(theme_path):         
            with open(theme_path) as json_file:
              grid_props = json.load(json_file)  


      colors = grid_props.get('colors',None)
      if colors is not None:
        
        self.base_color = colors.get('base', self.base_color)
        self.grid_color = colors.get('lines', self.grid_color)
        self.start_color = colors.get('start', self.start_color)
        self.start_text_color = colors.get('start_text', self.start_text_color)
        self.exit_color = colors.get('exit', self.exit_color)
        self.exit_text_color = colors.get('exit_text', self.exit_text_color)
        self.border_color = colors.get('border', self.border_color)
        self.wall_color = colors.get('walls', self.wall_color)   

      widths = grid_props.get('widths',None)
      if widths is not None:
        self.padding = widths.get('padding', self.padding)
        self.wall_width = widths.get('walls', self.wall_width)     
        self.border_width = widths.get('border', self.border_width)     


  def calculate_dimensions(self):
    ' calculate dimensions of the canvases in pixels '
    self.width_pixels = self.grid.width * self.cell_pixels + (self.padding*2)
    self.height_pixels = self.grid.height * self.cell_pixels + (self.padding*2)   
    self.total_width = self.width_pixels
    self.total_height = self.height_pixels    

    # if a compass or info side panel are being added expand the width
    if self.add_compass or (self.side_panel is not None): 
      # test if a width has been specified for the panel
      if type(self.side_panel) == int:
        self.total_width += self.side_panel
      elif type(self.side_panel) == dict:        
        # side panel has been specified as a dictionary
        self.total_width += self.side_panel.get('width',100)
        # self.total_height += self.side_panel.get('height',0)
      else:
        # create the side panel with the default width
        self.total_width += 100

    # if a bottom panel is specified increase the height    
    if self.bottom_panel is not None: 
      # test if a height has been specified for the panel
      if type(self.bottom_panel) == int:
        self.total_height += self.bottom_panel
      elif type(self.bottom_panel) == dict:
        # bottom panel has been specified as a dictionary
        # e.g. bottom_panel':{'width':200,'height':50,'color':'#644242'}        
        # self.total_width += self.bottom_panel.get('width',0)
        self.total_height += self.bottom_panel.get('height',50)
      else:
        # create the side panel with the default height      
        self.total_height += 50
    
    # calculate the number of pixels to center of a square
    self.center = self.cell_pixels//2 - self.padding     
    
    
  def create_canvases(self):      
    # calculate cell values in pixels
    self.calculate_dimensions()    
                
    # create the canvas layers
    self.canvases = MultiCanvas(n_canvases=self.num_canvases, 
                                width=self.total_width, 
                                height=self.total_height, 
                                sync_image_data=True)     


  '''
      Helper Functions
  '''                 

  def grid_to_pixels( self, grid_pos, xoff = 0, yoff = 0 ):
    x = (grid_pos[0] * self.cell_pixels) + self.padding + xoff
    y = (grid_pos[1] * self.cell_pixels) + self.padding + yoff   
    return x,y   


  def get_center(self,x,y):
    ''' get the center of the tile '''
    cx = x + self.center
    cy = y + self.center      
    return cx,cy    


  '''
      Draw Functions
  '''

  def draw_rect(self, canvas_index, width, height, color, x=0, y=0):
    ''' draw a rectangle of the supplied size and color '''
    canvas = self.canvases[canvas_index]
    canvas.fill_style = color
    canvas.fill_rect(x, y, width, height) 


  # def draw_area_walls(self):
  #   pass

  def set_reward_area(self,x,y,wd,ht,reward):
    pass

  def draw_grid_areas(self):
    ''' draw any grid areas 
      - these are areas on the grid that can be moved to '''
    for area in self.grid.grid_areas:
      try:  
        x,y,wd,ht = self.grid.get_area_defn(area[0])
        width  = wd * self.cell_pixels 
        height = ht * self.cell_pixels 
        px, py = self.grid_to_pixels([x,y])

        canvas_index = Level.Base     
        self.draw_rect( canvas_index, width, height, area[1], px, py )                

      except:
        # ignore bad entries
        pass   


  def draw_base_areas(self):
    ''' draw any base areas 
      - these are areas off the grid that cannot be moved to '''
    for area in self.grid.base_areas:
      try:  

        area_only = False        
        # test if only the area defn has been supplied
        if type(area[0]).__name__ == 'int':
          x,y,wd,ht = self.grid.get_area_defn(area)
          area_only = True
        else:
          x,y,wd,ht = self.grid.get_area_defn(area[0])
        width  = wd * self.cell_pixels 
        height = ht * self.cell_pixels 
        px, py = self.grid_to_pixels([x,y])

        # adjust the area if its at the edges
        half_border = self.border_width//2
        if x == 0:
          px -= half_border
          width += half_border
        if y == 0:
          py -= half_border
          height += half_border
        if (x+wd) == self.grid.width:          
          width += half_border          
        if (y+ht) == self.grid.height:          
          height += half_border                

        # draw the area
        color = "white" # base areas are white by default
        if area_only == False and len(area) > 1: 
          if type(area[1]).__name__ == 'str': color = area[1]        
        canvas_index = Level.Grid     
        self.draw_rect( canvas_index, width, height, color, px, py )


        #
        # Draw Borders
        # 
        
        borders = None  # all borders are added by default
        if area_only == False and len(area) > 1: 
          if type(area[1]).__name__ != 'str': borders = area[1]
          if len(area) == 3: borders = area[2]

        if borders == None:
          # base areas by default have their borders drawn
          # - so if no borders have been defined by the setup 
          # add all borders that are not on the edges of the canvas
          borders = []
          if y != 0: borders.append(('N'))
          if (y+ht) != self.grid.height: borders.append(('S'))
          if x != 0: borders.append(('W'))
          if (x+wd) != self.grid.width: borders.append(('E'))        

        canvas = self.canvases[Level.Grid]
        canvas.set_line_dash([0,0])
        canvas.line_cap = 'square'

        for border in borders:                      
          edge = border[0]
          canvas.stroke_style = border[1] if len(border) > 1 else self.border_color                                                          
          canvas.line_width = border[2] if len(border) == 3 else self.border_width 
            
          # top border
          if edge == 'N':
            x1 = self.padding + px
            y1 = self.padding + py 
            x2 = x1 + width-(2*self.padding)
            y2 = y1
            self.draw_line( canvas, x1, y1, x2, y2 )

          # bottom border          
          if edge == 'S':
            x1 = self.padding + px
            y1 = py + height - self.padding 
            x2 = x1 + width-(2*self.padding)
            y2 = y1
            self.draw_line( canvas, x1, y1, x2, y2 )

          # left border          
          if edge == 'W':
            x1 = self.padding + px
            y1 = self.padding + py 
            x2 = x1
            y2 = y1 + height-(2*self.padding)
            self.draw_line( canvas, x1, y1, x2, y2 )        

          # right border
          # if edge == 'E' and (x+wd) != self.grid.width:
          if edge == 'E':
            x1 = px + width - self.padding
            y1 = self.padding + py 
            x2 = x1
            y2 = y1 + height-(2*self.padding)
            self.draw_line( canvas, x1, y1, x2, y2 )                                      

      except:
        # ignore bad entries
        pass        


  def draw_line( self, canvas, x1, y1, x2, y2 ):
      ''' draw a straight line on the canvas '''
      canvas.begin_path()
      canvas.move_to(x1 , y1 )
      canvas.line_to(x2 , y2 )
      canvas.stroke()      


  # def draw_areas(self,canvas_index=0):
  #   ''' change the base color for any specified areas '''    
  #   for area in self.base_areas:
  #     try:  
  #       x,y,*args = area[0]  
  #       wd,ht = args if args else (1,1)
        
  #       # by default draw areas to the base canvas unless another index
  #       # is specified
  #       area_canvas = area[2] if len(area) >= 3 else Level.Base

  #       width  = wd * self.cell_pixels 
  #       height = ht * self.cell_pixels 
  #       px, py = self.grid_to_pixels([x,y])
        
  #       if canvas_index == Level.Base and canvas_index == area_canvas:

  #         self.draw_rect( canvas_index, width, height, area[1], px, py )  
        
  #       elif canvas_index > Level.Base:
  #         # draw area borders if specified

  #         half_border_width = math.floor(self.border_width / 2)

  #         ax = px
  #         ay = py
  #         aht = height
  #         awd = width


  #         # adjust the area at the edges to compensate for border
  #         if y == 0: 
  #           ay = py - half_border_width
  #           aht += half_border_width            
          
  #         if (y+ht) == self.grid.height: 
  #           ay = py + (half_border_width-1)
  #           aht += half_border_width        

  #         if x == 0: 
  #           ax = px - half_border_width
  #           awd += half_border_width            

  #         if (x+wd) == self.grid.width: 
  #           ax = px + (half_border_width-1)
  #           awd += half_border_width                     

  #         self.draw_rect( area_canvas, awd, aht, area[1], ax, ay )   

  #         if canvas_index == Level.Underlay:        

  #           if len(area) == 4: 

  #             canvas = self.canvases[Level.Grid]
  #             canvas.set_line_dash([0,0])
  #             canvas.line_cap = 'square'    

  #             for border in area[3]:              
                
  #               edge = border[0]
  #               canvas.stroke_style = border[1] if len(border) > 1 else self.border_color                                                
  #               line_width = border[2] if len(border) == 3 else self.border_width                     
  #               canvas.line_width = line_width
                
  #               def draw_wall( x1,y1,x2,y2 ):
  #                   canvas.begin_path()
  #                   canvas.move_to(x1 , y1 )
  #                   canvas.line_to(x2 , y2 )
  #                   canvas.stroke()

  #               if   edge == 'N': draw_wall( px, py, px+width, py ) 
  #               elif edge == 'S': draw_wall( px, py+height, px+width, py+height ) 
  #               elif edge == 'E': draw_wall( px+width, py, px+width, py+height ) 
  #               elif edge == 'W': draw_wall( px, py, px, py+height )                                   

  #     except:
  #       # ignore bad entries
  #       pass      


  def draw_grid(self,canvas):        
    ''' add dashed lines showing grid '''             
    # canvas.clear()
    canvas.stroke_style = self.grid_color
    canvas.line_width = 1
    canvas.set_line_dash([4,8])    

    # draw the grid onto the canvas
    for y in range(self.grid.height):   
      for x in range(self.grid.width):   
        canvas.stroke_rect(self.cell_pixels * x + self.padding, 
                           self.cell_pixels * y + self.padding, 
                           self.cell_pixels,
                           self.cell_pixels)
      
      
  def draw_start(self,canvas):      
    ''' add the start '''    
    start_x, start_y = self.grid_to_pixels( self.grid.start )
    canvas.fill_style = self.start_color
    canvas.fill_rect(start_x, start_y, self.cell_pixels, self.cell_pixels)       
    canvas.fill_style = self.start_text_color
    
    if self.show_start_text:
      canvas.text_align = 'left'
      canvas.font = 'bold 17px sans-serif'
      canvas.fill_text(str("START"), start_x + 5, start_y + 38)     
      
      
  def draw_exit(self, canvas):
    ''' add the exit '''    
    end_x, end_y = self.grid_to_pixels( self.grid.end )
    canvas.fill_style = self.exit_color
    canvas.fill_rect(end_x, end_y, self.cell_pixels, self.cell_pixels)    
    canvas.fill_style = self.exit_text_color
    if self.show_end_text:
      canvas.text_align = 'left'
      canvas.font = 'bold 20px sans-serif'
      canvas.fill_text(str("EXIT"), end_x + 10, end_y + 40)       
      
      
  def draw_border(self,canvas):
    ''' draw the level border '''     
    canvas.stroke_style = self.border_color
    canvas.line_width = self.border_width
    canvas.set_line_dash([0,0])
    canvas.stroke_rect(self.padding, 
                       self.padding, 
                       self.width_pixels-(2*self.padding),
                       self.height_pixels-(2*self.padding))     


  def draw_maze(self,canvas):    
    ''' draw any maze or walls to the canvas '''
    if self.grid.add_maze: 
      self.grid.maze.write_to_canvas( canvas,
                                      self.grid.height*self.cell_pixels,
                                      self.padding, 
                                      color = self.wall_color,
                                      wall_width = self.wall_width)    

  '''
      Puddles
  '''

  def load_puddle_sprite(self):        
      ' load the puddle sprite image and when loaded callback to split it into individual sprites '   
      image_path = os.path.join(self.grid.working_directory,'images/big_puddle.png')
      self.big_puddle = Image.from_file(image_path)     
      
      if self.grid.drawmode == 'colab':
        # load a small puddle sprite
        image_path = os.path.join(self.grid.working_directory,'images/small_puddle.png')
        self.small_puddle = Image.from_file(image_path)    
      else:
        # create a canvas from the big puddle sprite
        self.puddle_canvas = Canvas(width=self.cell_pixels, height=self.cell_pixels, sync_image_data=True)                 
        self.puddle_canvas.draw_image( self.big_puddle, 0, 0 )          



  def draw_puddles(self):  
    ''' draw the list of puddles onto the canvas '''   

    # test if puddles have been defined
    if self.grid.puddles:      
      canvas = self.canvases[Level.Grid]
      with hold_canvas(canvas):     
        for (x, y), puddle_size in self.grid.puddles:
          self.draw_splash( canvas, x, y, puddle_size )    


  def draw_splash(self,canvas,x,y,puddle_type):
    ''' draw the specified puddle size at the given location '''
    if puddle_type > 0:

      # create a puddle canvas, containing the scaled and randomly rotated
      # puddle - not current supported on colab
      if self.grid.drawmode != 'colab':

        # scale the puddle image according to its type (big or small)
        scale = puddle_type / 2

        # create a new canvas for each splash
        splash_canvas = Canvas(width=self.cell_pixels, height=self.cell_pixels)
        with hold_canvas(splash_canvas):          

            pos_x = self.cell_pixels//2
            pos_y = self.cell_pixels//2

            # Choose a random rotation angle 
            # (but first set the rotation center with `translate`)
            splash_canvas.translate(pos_x, pos_y)
            splash_canvas.rotate(uniform(0., pi))

            # scale the image
            splash_canvas.scale(scale)

            # Restore the canvas center
            splash_canvas.translate( -pos_x, -pos_y )

            # Draw the sprite          
            splash_canvas.draw_image(self.puddle_canvas, 0, 0)          

      x_px = x * self.cell_pixels + self.padding
      y_px = y * self.cell_pixels + self.padding
      if self.grid.drawmode == 'colab':
        if puddle_type==1:
          canvas.draw_image(self.small_puddle,x_px,y_px,width=self.cell_pixels,height=self.cell_pixels)
        else:
          canvas.draw_image(self.big_puddle,x_px,y_px,width=self.cell_pixels,height=self.cell_pixels)
      else:
        canvas.draw_image(splash_canvas,x_px,y_px,width=self.cell_pixels,height=self.cell_pixels)


  def draw_compass(self,canvas):      
    ''' draw the compass '''
    
    if self.add_compass: 
      arrows = Arrows(64,2,length=15,width=5,height=11)
      arrows.draw(canvas,
                  self.width_pixels + 27, 14,
                  [Direction.North,Direction.West,Direction.South,Direction.East],
                  center_width = 28 )

      canvas.font = 'bold 20px sans-serif'
      canvas.fill_text(str("W"), self.width_pixels + 13, 52)    
      canvas.fill_text(str("N"), self.width_pixels + 49, 18)     
      canvas.fill_text(str("E"), self.width_pixels + 82, 52)     
      canvas.fill_text(str("S"), self.width_pixels + 51, 85)        


  def draw_info_panel(self):
      ''' add any background color for the info panels '''     

      if type(self.side_panel) == dict:
        # e.g. side_panel':{'width':200,'height':50,'color':'#644242'}
        width = self.side_panel.get('width',200)
        height = self.side_panel.get('height',self.height_pixels)
        color = self.side_panel.get('color','white')
        self.draw_rect(Level.Base, width, height, color, x = self.width_pixels, y = 0)   

      if type(self.bottom_panel) == dict:
        # e.g. bottom_panel':{'width':200,'height':50,'color':'#644242'}
        width = self.bottom_panel.get('width',self.total_width)
        height = self.bottom_panel.get('height',100)
        color = self.bottom_panel.get('color','white')                
        self.draw_rect(Level.Base, width, height, color, x = 0, y = self.height_pixels)     


  def clear( self, all_info = False ):
    ''' clear anything currently in the info panels '''    

    canvas = self.canvases[Level.Overlay]
        
    if self.side_panel is not None:
      canvas.clear_rect(self.width_pixels,0,(self.total_width-self.width_pixels),self.total_height)   

    if self.bottom_panel is not None:
      canvas.clear_rect(0,self.height_pixels,self.total_width,(self.total_height-self.height_pixels))         

    if all_info == True:
      canvas.clear()


  '''
      Main Draw Routine
  '''



  def draw_level(self):
    ''' draw the base of the grid ''' 

    self.canvases[Level.Overlay].clear()
    self.canvases[Level.Grid].clear()

    # do any info panel setup
    self.draw_info_panel() 

    # put the coloured rectangle on the base layer    
    self.draw_rect(Level.Base, self.width_pixels, self.height_pixels, self.base_color)   

    # change the color of any areas that have been specified as grid level
    self.draw_grid_areas()         
        
    canvas = self.canvases[Level.Grid]
    
    # change after ipycanvas v11?
    with hold_canvas(canvas):        
      self.draw_start(canvas)
      self.draw_exit(canvas)   
      self.draw_grid(canvas)
      # self.draw_areas(canvas_index=Level.Grid)         
      self.draw_maze(canvas) 
      self.draw_border(canvas)        
      # self.draw_areas(canvas_index=Level.Underlay) 
      self.draw_base_areas()        
      self.draw_compass(canvas) 

    # draw puddles separately as may require loading of a sprite
    self.draw_puddles()         


  # test function
  def create_canvas(self):
    
    width = 200
    height = 200

    # Create a multi-layer canvas with 2 layers
    canvas = MultiCanvas(2, width=width, height=height, sync_image_data=True)

    # color of the grid base layer
    canvas[0].fill_style = 'orange' 
    canvas[0].fill_rect(0, 0, width, height) 

    return canvas                                       


import gym
import numpy as np

class Dynamic(gym.Space):

  def __init__(self, action_list = []):
      ' set the list of initially available actions '      
      self.set_actions(action_list)
      
  def sample(self):
      ' select a random action from the set of available actions '
      return np.random.choice(self.available_actions)    
    
  def set_actions(self,actions):
      self.available_actions = actions
      self.n = len(actions)    
      
  def get_available_actions(self):
      return [str(action) for action in self.available_actions] 



import os
from enum import IntEnum
from .maze import Maze
import numpy as np
from typing import Union


class Puddle(IntEnum):
    Dry, Small, Large = range(3)   


class GridBase():
 
  maze = None                # instance of maze if defined  
  puddles = None             # set of tiles where puddles exist
  base_areas = []            # any areas that exist on the base 
  grid_areas = []            # and areas that exist on the grid
  
  debug_maze = False         # write the maze to a svg file  

  grid_rewards = []          # the rewards calculated for the whole grid

  
  def __init__( self, working_directory: str = ".", **kwargs: dict ):
    
    self.working_directory = working_directory
        
    # test if a special drawing mode should be used
    # - currently required to run on Google Colab
    # - try to automatically detect if running in Colab using environ
    self.drawmode = kwargs.get('drawmode', 'colab' if 'COLAB_GPU' in os.environ else "" )

    self.width = kwargs.get('width',3)
    self.height = kwargs.get('height',3)    
    
    # the start and end positions in the grid
    # - by default these are the top-left and bottom-right respectively
    self.start = kwargs.get('start',[0,0])       
    self.end = kwargs.get('end',[self.width-1,self.height-1])    

    # setup any puddles    
    self.puddles = kwargs.get('puddles',None)     

    # setup up any properties defined for the puddles
    puddle_props = kwargs.get('puddle_props',{})
    self.large_puddle_reward = puddle_props.get('large_reward',-4)
    self.small_puddle_reward = puddle_props.get('small_reward',-2)
    self.large_puddle_probability = puddle_props.get('large_prob',0.4)
    self.small_puddle_probability = puddle_props.get('small_prob',0.6)    

    # setup any base-level areas    
    self.base_areas = kwargs.get('base_areas',[])     

    # setup any grid-level areas    
    self.grid_areas = kwargs.get('grid_areas',[])       

    # setup any maze and walls
    self.add_maze = kwargs.get('add_maze',False)
    self.maze_seed = kwargs.get('maze_seed',0)
    self.make_maze()
    self.toggle_walls( kwargs.get('walls',[]) )

    # calculate the rewards for each cell in the grid
    self.grid_rewards = self.get_reward()
    

  '''
      Maze and Walls
  '''

  def make_maze(self):
    if self.add_maze:
      if self.maze is None:
        self.maze = Maze(self.width, self.height, self.start[0], self.start[1], seed = self.maze_seed)
        self.maze.make_maze()        
      if self.debug_maze: 
        self.maze.write_svg(os.path.join(self.working_directory, "maze.svg"))


  def toggle_walls(self, walls):
    ''' add or remove the specified walls from the grid '''

    # if a maze isnt already defined begin with a maze with no walls
    if self.maze is None:
      self.maze = Maze(self.width, self.height, self.start[0], self.start[1], no_walls = True)
      self.add_maze = True # we now have a maze to add to the canvas

    for (x, y), direction in walls:
        current_cell = self.maze.cell_at(x,y)
        if   direction == 'E': next_cell = self.maze.cell_at(x+1,y)
        elif direction == 'W': next_cell = self.maze.cell_at(x-1,y)
        elif direction == 'N': next_cell = self.maze.cell_at(x,y-1)
        elif direction == 'S': next_cell = self.maze.cell_at(x,y+1)
        
        # add a new wall if none already otherwise remove
        current_cell.toggle_wall(next_cell, direction)           


  '''
      Puddles
  '''        

  def get_puddle_size( self, x, y ):
    ''' get the size of the puddle at the supplied location '''
    if self.puddles is not None:
      for (px,py),puddle_size in self.puddles:
        if x==px and y==py:         
          return Puddle(puddle_size)   

    return Puddle.Dry    


  def get_transition_probability( self, x, y ):
    ''' get the probability of moving to the target state when starting in the state at (x,y) '''
    puddle_size = self.get_puddle_size( x, y )         

    if puddle_size == Puddle.Large: return self.large_puddle_probability
    if puddle_size == Puddle.Small: return self.small_puddle_probability
    
    # if no puddle then guaranteed to reach target
    return 1.     


  def get_reward( self, x: int = None, y: int = None ) -> Union[int,np.ndarray]:
    ''' return the reward for the specified grid cell '''
    if (x is None) or (y is None):
      return self.get_reward_array()
    else:
      return self.get_reward_value(x,y)


  def get_reward_array(self) -> np.ndarray:
    ''' return a numpy array containing the reward value for all grid cells '''

    if len(self.grid_rewards) == 0:       
      height = self.height
      width = self.width
      reward_arr = np.zeros((height,width)).astype(int)
      for y in range(height):
        for x in range(width):
          reward_arr[y][x] = self.get_reward_value(x,y) 
      return reward_arr      

    # grid rewards already calculated
    return self.grid_rewards
        

  def get_reward_value( self, x, y ):
    ''' return the reward obtained for moving to the specified state 

        The amount of reward is a function of the puddle size:
        - no puddle = -1
        - small puddle = -2
        - large puddle = -4

        This represents the amount of time required to move through the puddle
    '''

    if len(self.grid_rewards) > 0: 
      # use the pre-calculated rewards
      # - note this holds the rewards as row,col so must swap x and y   
      return self.grid_rewards[y,x]

    # the terminal state has a reward of zero
    if (x == self.end[0]) and (y == self.end[1]):
      return 0

    puddle_size = self.get_puddle_size( x, y )
    if   puddle_size == Puddle.Large: return self.large_puddle_reward
    elif puddle_size == Puddle.Small: return self.small_puddle_reward  

    # no rewards exist off the grid
    for area in self.base_areas: 
      # test if only the area defn has been supplied
      if type(area[0]).__name__ == 'int':
        ax,ay,aw,ah = self.get_area_defn(area)
      else:
        ax,ay,aw,ah = self.get_area_defn(area[0])      
      if self.in_area(x,y,ax,ay,aw,ah):
        return 0

    # if any grid areas exist these can set different rewards
    # - the most recently defined area is the one whose reward will be taken
    # for a cell
    cell_reward = -1
    for area in self.grid_areas:      
      if len(area) > 2:   
        try:     
          ax,ay,aw,ah = self.get_area_defn(area[0]) 
          if self.in_area( x,y,ax,ay,aw,ah ):
            cell_reward = area[2]        
        except:
          pass

    return cell_reward    

  '''
     Areas
  '''

  def get_area_defn( self, area):
    ''' extract the area properties from the supplied area tuple '''
    x,y,*args = area
    wd,ht = args if args else (1,1) # default to unit-square if no values specified
    return x,y,wd,ht


  def in_area(self,x,y,ax,ay,aw,ah):
    if (x >= ax and x < (ax + aw)) and ((y >= ay and y < (ay + ah))):
      return True
    return False    


import numpy as np
import gym
from gym.spaces import Discrete, MultiDiscrete


class BabyRobotEnv_v1(gym.Env):
  
    def __init__(self, **kwargs):
        super().__init__()

        # dimensions of the grid
        self.width = kwargs.get('width',3)
        self.height = kwargs.get('height',3)      
      
        # define the maximum x and y values
        self.max_x = self.width - 1
        self.max_y = self.height - 1

        # there are 5 possible actions: move N,E,S,W or stay in same state
        self.action_space = Discrete(5)          

        # the observation will be the coordinates of Baby Robot            
        self.observation_space = MultiDiscrete([self.width, self.height])
                                        
        # Baby Robot's position in the grid
        self.x = 0
        self.y = 0

    def step(self, action):                
        obs = np.array([self.x,self.y])        
        reward = -1            
        done = True
        info = {}
        return obs, reward, done, info

    def reset(self):
        # reset Baby Robot's position in the grid
        self.x = 0
        self.y = 0                
        return np.array([self.x,self.y])
          
    def render(self, mode):
        pass


import gym
import numpy as np
from gym.spaces import Discrete, MultiDiscrete

from .lib.direction import Direction
from .lib.actions import Actions
from .lib.dynamic_space import Dynamic
from .lib.grid_level import GridLevel
from .lib.robot_draw import RobotDraw


class BabyRobotEnv_v7(gym.Env):
    
    def __init__(self, **kwargs):
        super().__init__()

        # dimensions of the grid
        self.width = kwargs.get('width',3)
        self.height = kwargs.get('height',3)      
      
        # define the maximum x and y values
        self.max_x = self.width - 1
        self.max_y = self.height - 1 

        # initially no actions are available      
        self.dynamic_action_space = Dynamic()   

        # by default use a dynamic action space 
        if kwargs.get('action_space','dynamic') == 'dynamic':
          self.action_space = self.dynamic_action_space              
        else:
          # use discrete action space 
          # - required for Stable Baselines environment checker which cant yet
          #   recognise dynamic spaces
          # - all actions are available in each state
          # - there are 5 possible actions: move N,E,S,W or stay in same state
          self.action_space = Discrete(5)        


        # the observation will be the coordinates of Baby Robot            
        self.observation_space = MultiDiscrete([self.width, self.height])                       

        # the start and end positions in the grid
        # - by default these are the top-left and bottom-right respectively
        self.start = kwargs.get('start',[0,0])       
        self.end = kwargs.get('end',[self.max_x,self.max_y])        
        
        # Baby Robot's initial position
        # - by default this is the grid start 
        self.initial_pos = kwargs.get('initial_pos',self.start)  

        # Baby Robot's position in the grid
        self.x = self.initial_pos[0]
        self.y = self.initial_pos[1]        

        # graphical creation of the level
        self.level = GridLevel( **kwargs )  
        
        # add baby robot
        self.robot = RobotDraw(self.level,**kwargs)   
        self.robot.draw()          

        # set the initial position and available actions
        self.reset() 


    #
    # Helper Methods
    #         

    def take_action(self, action):
        ''' apply the supplied action 

            returns:
            - the reward obtained for taking the action
            - a flag to indicate if the target state was reached - if it wasn't this indicates
              that a slip has occurred
        '''         

        # convert the action into a direction bitfield 
        direction = Direction.from_action(action)  
          
        # calculate the postion of the next state and the reward for moving there
        next_pos,reward,target_reached = self.level.get_next_state( self.x, self.y, direction )  

        # store the new position
        self.x = next_pos[0]
        self.y = next_pos[1]
    
        # update the available actions for the new position        
        self.set_available_actions()      
        return reward, target_reached  


    def get_available_actions( self ):
        ''' test which actions are allowed at the specified grid state '''      

        # get the available actions from the grid level
        direction_value = self.level.get_directions(self.x,self.y) 

        # convert the grid directions into environment actions
        action_list = []       
        if direction_value & Direction.North: action_list.append( Actions.North )
        if direction_value & Direction.South: action_list.append( Actions.South )
        if direction_value & Direction.East:  action_list.append( Actions.East ) 
        if direction_value & Direction.West:  action_list.append( Actions.West )                
        return action_list     

              
    def set_available_actions( self ):
        ' set the list of available actions into the action space '
        action_list = self.get_available_actions()   
        self.dynamic_action_space.set_actions( action_list )      


    def show_available_actions( self ):
        ''' return a string of avaiable actions for current state '''
        available_actions = str(self.dynamic_action_space.get_available_actions()).replace("'","")
        return f"({self.x},{self.y}) {available_actions:36}"


    def show_info(self,info):
        ''' display the supplied information on the grid level '''
        self.level.show_info( info )


    def clear_info(self,all_info=False):
        ''' clear any current information of the grid level '''
        self.level.clear(all_info)         


    #
    # Gym Interface Methods
    #

    def step(self, action): 
        ''' take the action and update the position '''        
        reward, target_reached = self.take_action(action)      
        obs = np.array([self.x,self.y])      
            
        # set the 'done' flag if we've reached the exit
        done = (self.x == self.end[0]) and (self.y == self.end[1])
          
        info = {'target_reached':target_reached}
        return obs, reward, done, info 


    def render(self, mode='human', info=None ):                 
        ''' render as an HTML5 canvas '''
              
        # move baby robot to the current position
        self.robot.move(self.x,self.y) 

        # write the info to the grid side-panel      
        self.level.show_info(info) 
        return self.level.draw()  


    def reset(self):
        ''' reset Baby Robot's position in the grid '''
        self.robot.set_cell_position(self.initial_pos)      
        self.robot.reset()               
        self.x = self.initial_pos[0]
        self.y = self.initial_pos[1]        
        self.set_available_actions()
        return np.array([self.x,self.y])  




from .actions import Actions
from .direction import Direction
from .arrows import Arrows

from .grid_base import GridBase
from .grid_level import GridLevel
from .draw_grid import DrawGrid
from .robot_position import RobotPosition
from .robot_draw import RobotDraw





from enum import IntEnum

''' simple helper class to enumerate actions in the grid levels '''
class Actions(IntEnum):  
    Stay  = 0    
    North = 1
    East  = 2
    South = 3
    West  = 4

    # get the enum name without the class
    def __str__(self): return self.name  


from enum import IntFlag
from typing import List,Union

from babyrobot.envs.lib.actions import Actions

''' simple helper class to enumerate directions in the grid levels '''
class Direction(IntFlag):
    Stay = 0
    North = 1
    East = 2
    South = 4
    West = 8
    All = 15
        
    # get the enum name without the class
    def __str__(self): return self.name  


    def get_value( direction_list: List[int] ) -> int:
      ''' convert a list of directions into a single bitfield value '''
      dir_value = 0
      for direction,v in direction_list.items():
        # test the action is allowed
        if v == True:                            
          # calculate the postion of the next state              
          if   direction == 'N': dir_value += Direction.North          
          elif direction == 'S': dir_value += Direction.South  
          elif direction == 'E': dir_value += Direction.East  
          elif direction == 'W': dir_value += Direction.West        
      return dir_value
      

    def get_string( direction: Union[int, List[int]] ) -> str:
      ''' convert a direction into a string of the form 'NSEW' 
          - direction can either be a list or bitfield value
      ''' 
      if type(direction) is list:
        return Direction.get_string_from_list( direction )
      else:
        return Direction.get_string_from_value( direction )
      

    def get_string_from_value( direction_value: int ) -> str:
      ''' convert a direction bitfield value into a string of the form 'NSEW' '''    
      dir_list = Direction.get_list(direction_value)   
      return Direction.get_string_from_list(dir_list)


    def get_string_from_list( direction_list: List[int] ) -> str:
      ''' convert a list of directions into a string of the form 'NSEW' '''      
      dir_string = ""
      for direction in direction_list:          
        if direction == Direction.North: dir_string += "N"      
        if direction == Direction.South: dir_string += "S"                  
        if direction == Direction.East:  dir_string += "E"                
        if direction == Direction.West:  dir_string += "W"   
      return dir_string


    def get_list( direction_value: int ) -> List[int]:
      ''' convert a direction bitfield value into a list of directions '''       
      dir_list = []      
      if direction_value & Direction.North: dir_list.append( Direction.North )
      if direction_value & Direction.South: dir_list.append( Direction.South )
      if direction_value & Direction.East:  dir_list.append( Direction.East ) 
      if direction_value & Direction.West:  dir_list.append( Direction.West )    
      return dir_list  


    def from_action( action_value: Actions):
      ''' convert an action value to a direction '''
      if Actions(action_value) is not Actions.Stay:              
        return Direction(pow(2, action_value-1))
      return Direction.Stay


# Create a maze using the depth-first algorithm described at
# https://scipython.com/blog/making-a-maze/
# Christian Hill, April 2017.

import random

class Cell:
    """A cell in the maze.
    A maze "Cell" is a point in the grid which may be surrounded by walls to
    the north, east, south or west.
    """

    # A wall separates a pair of cells in the N-S or W-E directions.
    wall_pairs = {'N': 'S', 'S': 'N', 'E': 'W', 'W': 'E'}

    def __init__(self, x, y, no_walls = False):
        """Initialize the cell at (x,y). At first it is surrounded by walls."""

        self.x, self.y = x, y

        if no_walls:
          self.walls = {'N': False, 'S': False, 'E': False, 'W': False }
        else:
          self.walls = {'N': True, 'S': True, 'E': True, 'W': True}

    def has_all_walls(self):
        """Does this cell still have all its walls?"""

        # check if all items in the list are true
        return all(self.walls.values())

    def knock_down_wall(self, other, wall):
        """Knock down the wall between cells self and other."""

        self.walls[wall] = False
        other.walls[Cell.wall_pairs[wall]] = False

    def add_wall(self, other, wall):
        """Add a wall between cells self and other."""

        self.walls[wall] = True
        other.walls[Cell.wall_pairs[wall]] = True  

    def toggle_wall(self, other, wall):
        if self.walls[wall]:
          self.knock_down_wall(other,wall)
        else:
          self.add_wall(other,wall)    


class Maze:
    """A Maze, represented as a grid of cells."""

    def __init__(self, nx, ny, ix=0, iy=0, seed = None, no_walls = False):
        """Initialize the maze grid.
        The maze consists of nx x ny cells and will be constructed starting
        at the cell indexed at (ix, iy).
        """

        # set the seed for random if want to produce consistent maze
        if seed is not None: random.seed(seed)
        
        self.nx, self.ny = nx, ny
        self.ix, self.iy = ix, iy
        self.maze_map = [[Cell(x, y, no_walls) for y in range(ny)] for x in range(nx)]

        if no_walls:
          self.add_boundary_walls()

    def add_boundary_walls(self):
        """ add walls around the outside of the level """
        for y in range(self.ny):
          for x in range(self.nx):
            if y == 0:
              self.cell_at(x, y).walls['N'] = True
            elif (y+1) == self.ny:
              self.cell_at(x, y).walls['S'] = True

            if x == 0:
              self.cell_at(x, y).walls['W'] = True
            elif (x+1) == self.nx:
              self.cell_at(x, y).walls['E'] = True              

    def cell_at(self, x, y):
        """Return the Cell object at (x,y)."""
        return self.maze_map[x][y]
    
    def dimensions(self):
        return self.nx, self.ny

    def __str__(self):
        """Return a (crude) string representation of the maze."""

        maze_rows = ['-' * self.nx * 2]
        for y in range(self.ny):
            maze_row = ['|']
            for x in range(self.nx):
                if self.maze_map[x][y].walls['E']:
                    maze_row.append(' |')
                else:
                    maze_row.append('  ')
            maze_rows.append(''.join(maze_row))
            maze_row = ['|']
            for x in range(self.nx):
                if self.maze_map[x][y].walls['S']:
                    maze_row.append('-+')
                else:
                    maze_row.append(' +')
            maze_rows.append(''.join(maze_row))
        return '\n'.join(maze_rows)

    def write_svg(self, filename):
        """Write an SVG image of the maze to filename."""

        aspect_ratio = self.nx / self.ny
        # Pad the maze all around by this amount.
        padding = 10
        # Height and width of the maze image (excluding padding), in pixels
        height = 500
        width = int(height * aspect_ratio)
        # Scaling factors mapping maze coordinates to image coordinates
        scy, scx = height / self.ny, width / self.nx

        def write_wall(ww_f, ww_x1, ww_y1, ww_x2, ww_y2):
            """Write a single wall to the SVG image file handle f."""

            print('<line x1="{}" y1="{}" x2="{}" y2="{}"/>'
                  .format(ww_x1, ww_y1, ww_x2, ww_y2), file=ww_f)

        # Write the SVG image file for maze
        with open(filename, 'w') as f:
            # SVG preamble and styles.
            print('<?xml version="1.0" encoding="utf-8"?>', file=f)
            print('<svg xmlns="http://www.w3.org/2000/svg"', file=f)
            print('    xmlns:xlink="http://www.w3.org/1999/xlink"', file=f)
            print('    width="{:d}" height="{:d}" viewBox="{} {} {} {}">'
                  .format(width + 2 * padding, height + 2 * padding,
                          -padding, -padding, width + 2 * padding, height + 2 * padding),
                  file=f)
            print('<defs>\n<style type="text/css"><![CDATA[', file=f)
            print('line {', file=f)
            print('    stroke: #000000;\n    stroke-linecap: square;', file=f)
            print('    stroke-width: 5;\n}', file=f)
            print(']]></style>\n</defs>', file=f)
            # Draw the "South" and "East" walls of each cell, if present (these
            # are the "North" and "West" walls of a neighbouring cell in
            # general, of course).
            for x in range(self.nx):
                for y in range(self.ny):
                    if self.cell_at(x, y).walls['S']:
                        x1, y1, x2, y2 = x * scx, (y + 1) * scy, (x + 1) * scx, (y + 1) * scy
                        write_wall(f, x1, y1, x2, y2)
                    if self.cell_at(x, y).walls['E']:
                        x1, y1, x2, y2 = (x + 1) * scx, y * scy, (x + 1) * scx, (y + 1) * scy
                        write_wall(f, x1, y1, x2, y2)
            # Draw the North and West maze border, which won't have been drawn
            # by the procedure above.
            print('<line x1="0" y1="0" x2="{}" y2="0"/>'.format(width), file=f)
            print('<line x1="0" y1="0" x2="0" y2="{}"/>'.format(height), file=f)
            print('</svg>', file=f)

    def find_valid_neighbours(self, cell):
        """Return a list of unvisited neighbours to cell."""

        delta = [('W', (-1, 0)),
                 ('E', (1, 0)),
                 ('S', (0, 1)),
                 ('N', (0, -1))]
        neighbours = []
        for direction, (dx, dy) in delta:
            x2, y2 = cell.x + dx, cell.y + dy
            if (0 <= x2 < self.nx) and (0 <= y2 < self.ny):
                neighbour = self.cell_at(x2, y2)
                if neighbour.has_all_walls():
                    neighbours.append((direction, neighbour))
        return neighbours

    def make_maze(self):
        # Total number of cells.
        n = self.nx * self.ny
        cell_stack = []
        current_cell = self.cell_at(self.ix, self.iy)
        # Total number of visited cells during maze construction.
        nv = 1

        while nv < n:
            neighbours = self.find_valid_neighbours(current_cell)

            if not neighbours:
                # We've reached a dead end: backtrack.
                current_cell = cell_stack.pop()
                continue

            # Choose a random neighbouring cell and move to it.
            direction, next_cell = random.choice(neighbours)
            current_cell.knock_down_wall(next_cell, direction)
            cell_stack.append(current_cell)
            current_cell = next_cell
            nv += 1
            
    def write_to_canvas(self, canvas, maze_height, maze_padding, color='#000', wall_width=4 ): 
        ' draw the maze onto the canvas '

        aspect_ratio = self.nx / self.ny
        # Pad the maze all around by this amount.
        padding = maze_padding
        
        # Height and width of the maze image (excluding padding), in pixels
        height = maze_height
        width = int(height * aspect_ratio)
        
        # Scaling factors mapping maze coordinates to image coordinates
        scy, scx = height / self.ny, width / self.nx

        canvas.line_width = wall_width
        canvas.line_cap = 'square'    
        
        canvas.stroke_style = color       
        canvas.set_line_dash([0,0])        

        def draw_wall( x1,y1,x2,y2 ):
            canvas.begin_path()
            canvas.move_to(x1 + padding, y1 + padding)
            canvas.line_to(x2 + padding, y2 + padding)
            canvas.stroke()

        # Draw the "South" and "East" walls of each cell, if present (these
        # are the "North" and "West" walls of a neighbouring cell in
        # general, of course).
        for x in range(self.nx):
            for y in range(self.ny):
                if self.cell_at(x, y).walls['S']:
                    x1, y1, x2, y2 = x * scx, (y + 1) * scy, (x + 1) * scx, (y + 1) * scy
                    draw_wall(x1, y1, x2, y2)
                if self.cell_at(x, y).walls['E']:
                    x1, y1, x2, y2 = (x + 1) * scx, y * scy, (x + 1) * scx, (y + 1) * scy
                    draw_wall(x1, y1, x2, y2)

        # Draw the North and West maze border, which won't have been drawn
        # by the procedure above.
        draw_wall(0, 0, 0, height)        
        draw_wall(0, 0, width, 0)


""" Setup script for BabyRobotEnv """

# Standard library imports
import pathlib

# Third party imports
from setuptools import find_packages, setup

# The directory containing this file
HERE = pathlib.Path(__file__).resolve().parent

# The text of the README file is used as a description
README = (HERE / "README.md").read_text()

# This call to setup() does all the work
setup(
    name="babyrobot",
    version="1.0.0",
    description="An OpenAI Gym Environment for BabyRobot",
    long_description=README,
    long_description_content_type="text/markdown",
    url="https://github.com/WhatIThinkAbout/BabyRobotGym",
    author="Steve Roberts",
    author_email="steve@steveroberts.name",
    license="MIT",
    classifiers=[
        "License :: OSI Approved :: MIT License",
        "Programming Language :: Python",
        "Programming Language :: Python :: 3",
    ],
    packages = find_packages(exclude=("tests",)),
    include_package_data=True,
    install_requires=['gym','ipycanvas==0.11']    
)



import numpy as np

from .baby_robot_env_v2 import BabyRobotEnv_v2

from .lib.grid_level import GridLevel
from .lib.robot_draw import RobotDraw
from .lib.actions import Actions


''' the first graphical environment '''
class BabyRobotEnv_v3( BabyRobotEnv_v2 ):    
  
  def __init__(self, **kwargs):
      super().__init__(**kwargs)
      
      # graphical creation of the level
      self.level = GridLevel( **kwargs )  
      
      # add baby robot
      self.robot = RobotDraw(self.level,**kwargs)   
      self.robot.draw()        
      
  def reset(self):
      # reset Baby Robot's position in the grid
      self.robot.set_cell_position(self.initial_pos)      
      self.robot.reset()               
      self.x = self.initial_pos[0]
      self.y = self.initial_pos[1]        
      return np.array([self.x,self.y])      
      
  def render(self, mode='human', action=0, reward=0 ):                
      ''' render as an HTML5 canvas '''
      print(f"{Actions(action): <5}: ({self.x},{self.y}) reward = {reward}")    
      
      # move baby robot to the current position
      self.robot.move(self.x,self.y) 
      return self.level.draw() 